@@ -32,13 +32,41 @@
  * By using a #GtkClutterEmbed widget is possible to build, show and
  * interact with a scene built using Clutter inside a GTK+ application.
  *
- * <warning><para>Though #GtkClutterEmbed is a #GtkContainer subclass,
- * it is not a real GTK+ container; #GtkClutterEmbed is required to
- * implement #GtkContainer in order to embed a #GtkWidget through the
- * #GtkClutterActor class. Calling gtk_container_add() on a #GtkClutterEmbed
- * will trigger an assertion. It is strongly advised not to override the
- * #GtkContainer implementation when subclassing
- * #GtkClutterEmbed.</para></warning>
+ * ## Event handling with GtkClutterEmbed
+ *
+ * Due to re-entrancy concerns, you should not use GTK event-related
+ * API from within event handling signals emitted by Clutter actors
+ * inside a #GtkClutterEmbed.
+ *
+ * Event-related API, like the GTK drag and drop functions, or the
+ * GTK grab ones, cause events to be processed inside the GDK event
+ * loop; #GtkClutterEmbed and the Clutter event loop may use those
+ * events to generate Clutter events, and thus emit signals on
+ * #ClutterActors. If you use the event-related signals of a
+ * #ClutterActor to call the GTK API, one of the two event loops
+ * will try to re-enter into each other, and either cause a crash
+ * or simply block your application.
+ *
+ * To avoid this behavior, you can either:
+ *
+ *  - only use GTK+ event handling signals to call event-related
+ *    GTK functions
+ *  - let the main loop re-enter, by calling event-related GTK
+ *    functions from within an idle or a timeout callback
+ *
+ * You should also make sure you're not using GTK widgets that call
+ * event-related GTK API, like the grab functions in a #GtkMenu, in
+ * response to Clutter actor events.
+ *
+ * ## Using GtkClutterEmbed as a container
+ *
+ * Though #GtkClutterEmbed is a #GtkContainer subclass, it is not a
+ * real GTK+ container; #GtkClutterEmbed is required to implement the
+ * #GtkContainer virtual functions in order to embed a #GtkWidget
+ * through the #GtkClutterActor class. Calling gtk_container_add()
+ * on a #GtkClutterEmbed will trigger an assertion. It is strongly
+ * advised not to override the #GtkContainer implementation when
+ * subclassing #GtkClutterEmbed, to avoid breaking internal state.
  */
 
 #ifdef HAVE_CONFIG_H
@@ -83,12 +111,6 @@
 #include <gdk/gdkwayland.h>
 #endif
 
-G_DEFINE_TYPE (GtkClutterEmbed, gtk_clutter_embed, GTK_TYPE_CONTAINER);
-
-#define GTK_CLUTTER_EMBED_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), GTK_CLUTTER_TYPE_EMBED, GtkClutterEmbedPrivate))
-
-static gint num_filter = 0;
-
 struct _GtkClutterEmbedPrivate
 {
   ClutterActor *stage;
@@ -109,6 +131,8 @@ struct _GtkClutterEmbedPrivate
 #endif
 };
 
+static gint num_filter = 0;
+
 enum
 {
   PROP_0,
@@ -116,6 +140,8 @@ enum
   PROP_USE_LAYOUT_SIZE
 };
 
+G_DEFINE_TYPE_WITH_PRIVATE (GtkClutterEmbed, gtk_clutter_embed, GTK_TYPE_CONTAINER)
+
 static void
 gtk_clutter_embed_send_configure (GtkClutterEmbed *embed)
 {
@@ -305,9 +331,9 @@ gtk_clutter_embed_ensure_subsurface (GtkClutterEmbed *embed)
   gtk_surface = gdk_wayland_window_get_wl_surface (gdk_window_get_toplevel(window));
 
   priv->subsurface =
-          wl_subcompositor_get_subsurface (priv->subcompositor,
-                                           priv->clutter_surface,
-                                           gtk_surface);
+    wl_subcompositor_get_subsurface (priv->subcompositor,
+                                     priv->clutter_surface,
+                                     gtk_surface);
 
   gtk_widget_translate_coordinates (widget, gtk_widget_get_toplevel (widget), 0, 0, &x, &y);
   gdk_window_get_origin (gtk_widget_get_parent_window (widget), &x, &y);
@@ -327,6 +353,13 @@ gtk_clutter_embed_realize (GtkWidget *widget)
   gint attributes_mask;
   gint border_width;
 
+#if defined(CLUTTER_WINDOWING_GDK)
+  if (clutter_check_windowing_backend (CLUTTER_WINDOWING_GDK))
+    {
+      GdkVisual *visual = clutter_gdk_get_visual ();
+      gtk_widget_set_visual (widget, visual);
+    }
+#endif
 #if defined(GDK_WINDOWING_X11) && defined(CLUTTER_WINDOWING_X11)
   if (clutter_check_windowing_backend (CLUTTER_WINDOWING_X11))
     {
@@ -1009,8 +1042,6 @@ gtk_clutter_embed_class_init (GtkClutterEmbedClass *klass)
   GtkContainerClass *container_class = GTK_CONTAINER_CLASS (klass);
   GParamSpec *pspec;
 
-  g_type_class_add_private (klass, sizeof (GtkClutterEmbedPrivate));
-
   gobject_class->dispose = gtk_clutter_embed_dispose;
   gobject_class->set_property = gtk_clutter_embed_set_property;
   gobject_class->get_property = gtk_clutter_embed_get_property;
@@ -1097,7 +1128,7 @@ gtk_clutter_embed_init (GtkClutterEmbed *embed)
   GtkClutterEmbedPrivate *priv;
   GtkWidget *widget;
 
-  embed->priv = priv = GTK_CLUTTER_EMBED_GET_PRIVATE (embed);
+  embed->priv = priv = gtk_clutter_embed_get_instance_private (embed);
   widget = GTK_WIDGET (embed);
 
   /* we have a real window backing our drawing */
